// icon to bitmap

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mem.h>
#include <dir.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef int BOOL;
#define TRUE 1
#define FALSE 0

const unsigned MAX_COMLINE_ARG_SIZE = 128;

const unsigned MAX_ICON_SIZE = 64;
const unsigned MED_ICON_SIZE = 32;
const unsigned MIN_ICON_SIZE = 16;

const unsigned MAX_COLOR_COUNT = 16;
const unsigned MED_COLOR_COUNT = 8;
const unsigned MIN_COLOR_COUNT = 2;

/*-----------------------------------------------------------------------------
Icon directory entry from icon file.
-----------------------------------------------------------------------------*/

static struct icon_directory_entry_tag
{                     // member must be
  BYTE width;         // MIN_ICON_SIZE, MED_ICON_SIZE or MAX_ICON_SIZE
  BYTE height;        // MIN_ICON_SIZE, MED_ICON_SIZE or MAX_ICON_SIZE
  BYTE color_count;   // MIN_COLOR_COUNT, MED_COLOR_COUNT or MAX_COLOR_COUNT
  BYTE reserved2;     // 0
  WORD planes;
  WORD bit_count;
  DWORD bytes_in_resource;
  DWORD image_offset;
} icon;

/*-----------------------------------------------------------------------------
Bitmap information block from icon file.
-----------------------------------------------------------------------------*/

static struct bitmap_info_tag
{                            // member must be
  DWORD size;                // sizeof(bminfo)
  DWORD width2;              // equal to width
  DWORD height2;             // equal to 2*height
  WORD planes2;              // anything
  WORD bits_per_pixel;       // 1 or 4
  DWORD compression;         // 0
  DWORD size_of_image;       // anything
  DWORD xpels;               // 0
  DWORD ypels;               // 0
  DWORD colors_used;         // 0
  DWORD important_colors;    // 0
} bminfo;

/*-----------------------------------------------------------------------------
RGB entry used in color table.
-----------------------------------------------------------------------------*/

struct RGB4
{
  BYTE blue;
  BYTE green;
  BYTE red;
  BYTE unused;
};

/*-----------------------------------------------------------------------------
Header in bitmap file.
-----------------------------------------------------------------------------*/

static struct bitmap_header_tag
{                            // member set to
  char type[2];              // "BM"
  DWORD file_size;
  WORD reserved1;            // 0
  WORD reserved2;            // 0
  DWORD bit_table_offset;
  DWORD info_header_size;    // sizeof(bminfo)
  DWORD width;
  DWORD height;
  WORD planes;               // 1
  WORD bits_per_pixel;       // 1 or 4
  DWORD compression;         // 0
  DWORD size_of_image;       // 0
  DWORD xpels;               // 0
  DWORD ypels;               // 0
  DWORD colors_used;         // 0
  DWORD important_colors;    // 0
} bitmap_header;

static char HELP[] =
  "ICO2BMP icon [-i n] [-b bitmap] [-n name] [-a] [-c red green blue] [-v]\n"
  "\n"
  "   icon    icon file specifications\n"
  "   n       index number of icon to be converted (first is number 0)\n"
  "   bitmap  bitmap file specifications\n"
  "   -a      append bitmap to existing bitmap file\n"
  "   name    bitmap name\n"
  "   red     red intensity of background (0-255)\n"
  "   green   green intensity of background (0-255)\n"
  "   blue    blue intensity of background (0-255)\n"
  "   -v      verbose output for diagnostic purposes\n";


static unsigned actual_color_count;
static RGB4 background_color;
static BOOL background_color_specified = FALSE;
static RGB4 colors[MAX_COLOR_COUNT];
static char icon_file_specifications[MAX_COMLINE_ARG_SIZE + 5];
static char bitmap_file_specifications[MAX_COMLINE_ARG_SIZE + 5];
static char bitmap_name[MAX_COMLINE_ARG_SIZE + 8];
static FILE *icon_file, *bitmap_file;
static unsigned scan_line_length;
static RGB4 pixel[MAX_ICON_SIZE * MAX_ICON_SIZE ];
static BOOL icon_number_specified = FALSE;
static unsigned icon_number = 0;
static BOOL rc_output;
static BYTE bitmap_buffer[16];
static unsigned bitmap_buffer_index = 0;
static BOOL verbose = FALSE;
static BOOL append = FALSE;
static BOOL bitmap_file_specified = FALSE;

/*-----------------------------------------------------------------------------
This function displays an appropriate error message and then exits with
errorlevel 1.
-----------------------------------------------------------------------------*/

static void error_exit(const char *format, ...)
{
  fprintf(stderr, "ICO2BMP: ");
  vfprintf(stderr, format, &format + 1);
  fprintf(stderr, "\n");
  exit(1);
}

/*-----------------------------------------------------------------------------
This function scans a number from the specified string. The number must be a
16-bit unsigned number, and the string may contain nothing else.
-----------------------------------------------------------------------------*/

static unsigned get_number(const char *s)
{
  unsigned long n = 0;
  const char *t;
  for (t = s; *t != 0; t++)
  {
    if ('0' <= *t && *t <= '9')
      n = 10 * n + *t - '0';
    else
      n = 0x10000L;
    if (n & 0xFFFF0000L)
      break;
  }
  if (n & 0xFFFF0000L)
    error_exit("Number out of range: %s", s);
  return (unsigned) n;
}

/*-----------------------------------------------------------------------------
This function scans a color intensity, which must be in the range 0-255.
-----------------------------------------------------------------------------*/

static unsigned get_color(const char *s)
{
  unsigned n = get_number(s);
  if (n > 255)
    error_exit("Color value out of range: %s", s);
  return n;
}

/*-----------------------------------------------------------------------------
This function finds the extension in the specified filespecs, and returns a
pointer to the period just before the extension. If the filespecs do not
contain an extension, it returns NULL.
-----------------------------------------------------------------------------*/

static char *extension(char *specs)
{
  char *t = specs + strlen(specs);
  while (t > specs)
  {
    int c = *--t;
    if (c == '.')
      return t;
    if (c == '\\' || c == ':')
      break;
  }
  return NULL;
}

/*-----------------------------------------------------------------------------
This function writes one byte of the bitmap file.
-----------------------------------------------------------------------------*/

static void write_bitmap(char c)
{
  if (putc(c, bitmap_file) == EOF)
    error_exit("Bitmap file write error");
}

/*-----------------------------------------------------------------------------
This function writes a string to the bitmap file.
-----------------------------------------------------------------------------*/

static void write_bitmap(char *s)
{
  while (*s != 0)
    write_bitmap(*s++);
}

/*-----------------------------------------------------------------------------
This function converts the LS 4 bits of its argument to a hexadecimal digit.
-----------------------------------------------------------------------------*/

static char hex(const unsigned c)
{
  int cc = c & 0xF;
  return cc < 10 ? cc + '0' : cc + ('A' - 10);
}

/*-----------------------------------------------------------------------------
This function flushes the bitmap buffer, when the bitmap is being written out
in .RC form.
-----------------------------------------------------------------------------*/

static void flush_bitmap_buffer(void)
{
  if (bitmap_buffer_index != 0)
  {
    unsigned i;
    write_bitmap("  '");
    for (i = 0; i < bitmap_buffer_index; i++)
    {
      write_bitmap(hex(bitmap_buffer[i] >> 4));
      write_bitmap(hex(bitmap_buffer[i]));
      if (i + 1 < bitmap_buffer_index)
        write_bitmap(' ');
    }
    write_bitmap("'\n");
    bitmap_buffer_index = 0;
  }
}

/*-----------------------------------------------------------------------------
This function writes a block of data to the bitmap file. If the bitmap is
being written out in .RC form, it writes the data to a buffer that is
converted to hexadecimal form and flushed when it is full.
-----------------------------------------------------------------------------*/

static void write_bitmap_block(void *buffer, const unsigned size)
{
  if (rc_output)
  {
    unsigned i;
    for (i = 0; i < size; i++)
    {
      bitmap_buffer[bitmap_buffer_index] = ((BYTE *) buffer)[i];
      if (++bitmap_buffer_index == sizeof(bitmap_buffer))
        flush_bitmap_buffer();
    }
  }
  else if (fwrite(buffer, size, 1, bitmap_file) != 1)
    error_exit("Bitmap file write error");
}

/*-----------------------------------------------------------------------------
Program execution starts here.
-----------------------------------------------------------------------------*/

int main(int argc, char **argv)
{

  // if there are no command-line arguments, display the help string and quit
  if (argc == 1)
  {
    puts(HELP);
    return 1;
  }

  // the first command-line argument is the icon file specifications
  strcpy(icon_file_specifications, argv[1]);

  // process other command-line arguments as options
  for (int i = 2; i < argc; i++)
  {
    if (strcmp(argv[i], "-b") == 0)
    {
      if (++i >= argc || argv[i][0] == '-')
        error_exit("Missing bitmap file specifications");
      strcpy(bitmap_file_specifications, argv[i]);
    }
    else if (strcmp(argv[i], "-a") == 0)
      append = TRUE;
    else if (strcmp(argv[i], "-i") == 0)
    {
      if (++i >= argc || argv[i][0] == '-')
        error_exit("Missing icon index number");
      icon_number_specified = TRUE;
      icon_number = get_number(argv[i]);
    }
    else if (strcmp(argv[i], "-n") == 0)
    {
      if (++i >= argc || argv[i][0] == '-')
        error_exit("Missing bitmap name");
      strcpy(bitmap_name, argv[i]);
    }
    else if (strcmp(argv[i], "-c") == 0)
    {
      if (++i >= argc || argv[i][0] == '-')
        error_exit("Missing background color value");
      background_color.red = get_color(argv[i]);
      if (++i >= argc || argv[i][0] == '-')
        error_exit("Missing background color value");
      background_color.green = get_color(argv[i]);
      if (++i >= argc || argv[i][0] == '-')
        error_exit("Missing background color value");
      background_color.blue = get_color(argv[i]);
      background_color_specified = TRUE;
    }
    else if (strcmp(argv[i], "-v") == 0)
      verbose = TRUE;
  }

  // add .ICO to icon file specifications, or check that the extension
  // is already .ICO
  {
    char *t = extension(icon_file_specifications);
    if (t == NULL)
      strcat(icon_file_specifications, ".ICO");
    else if (stricmp(t, ".ICO") != 0)
      error_exit("Icon file extension is not .ICO");
  }

  // if there are no bitmap file specifications, modify the icon file
  // specifications
  if (bitmap_file_specifications[0] == 0)
  {
    strcpy(bitmap_file_specifications, icon_file_specifications);
    strcpy(extension(bitmap_file_specifications), ".BMP");
  }
  // otherwise, add .BMP to the bitmap file specifications or check that
  // the extension is .BMP or .RC
  else
  {
    char *t = extension(bitmap_file_specifications);
    if (t == NULL)
      strcat(bitmap_file_specifications, ".BMP");
    else if (stricmp(t, ".RC") == 0)
      rc_output = TRUE;
    else if (stricmp(t, ".BMP") != 0)
      error_exit("Bitmap file extension is not .BMP or .RC");
  }

  // consistency of options
  if (append & !rc_output)
    error_exit("Append mode used only with .RC bitmap");

  // create bitmap name if the output is in .RC format
  if (rc_output)
  {
    if (bitmap_name[0] == 0)
      strcpy(bitmap_name, "BITMAP_!");
    for (char *t = bitmap_name; *t != 0; t++)
    {
      if (*t == '!')
      {
        char file_name[MAXFILE];
        fnsplit(icon_file_specifications, NULL, NULL, file_name, NULL);
        unsigned n = strlen(file_name);
        memmove(t + n, t + 1, strlen(t + 1));
        memcpy(t, file_name, n);
        break;
      }
    }
  }

  // display icon and bitmap file specifications, and also bitmap name
  // if there is one
  printf("%s  ->  %s", icon_file_specifications,
    bitmap_file_specifications);
  if (rc_output)
    printf(" (%s)", bitmap_name);
  printf("\n");

  // open the icon file
  icon_file = fopen(icon_file_specifications, "rb");
  if (icon_file == NULL)
    error_exit("Can't open icon file %s", icon_file_specifications);

  // read icon header
  {
    struct icon_header_tag
    {                            // member must be
      WORD reserved1;            // 0
      WORD type;                 // 1
      WORD number_of_icons;      // anything
    } header;
    if (fread(&header, sizeof(header), 1, icon_file) != 1)
      error_exit("Icon file does not contain a complete header");
    if (header.reserved1 != 0 || header.type != 1)
      error_exit("Icon file format error");
    if (!icon_number_specified && header.number_of_icons != 1)
    {
      error_exit("Icon file contains more than one icon\n"
        "             and icon number not specified");
    }
    if (icon_number > header.number_of_icons)
    {
      error_exit("Icon file does not contain icon number %ld\n",
        (unsigned long) icon_number);
    }
  }

  // read icon directory entry
  if (icon_number != 0 && fseek(icon_file,
    (long) icon_number * (long) sizeof(icon), SEEK_CUR) ||
    fread(&icon, sizeof(icon), 1, icon_file) != 1)
  {
    error_exit("Icon file format error");
  }

  // display icon dimension and number of colors
  printf("%d X %d, %d colors\n", icon.height, icon.width, icon.color_count);

  if (verbose)
  {
    printf("bitmap info header at %08lx\n", icon.image_offset     );
    printf("color table        at %08lx\n",
      icon.image_offset + sizeof(bminfo));
  }

  // a color count of 8 actually means 16 colors
  actual_color_count = icon.color_count == 2 ? 2 : 16;

  if (!background_color_specified)
  {
    background_color.red = background_color.green =
      background_color.blue = actual_color_count == 2 ? 255 : 192;
  }

  if (verbose)
  {
    printf("bitmap             at %08lx\n",
      icon.image_offset + sizeof(bminfo) + actual_color_count * sizeof(RGB4));
  }

  // check icon directory entry for validity
  if
  (
    icon.width                != MIN_ICON_SIZE &&
    icon.width                != MED_ICON_SIZE &&
    icon.width                != MAX_ICON_SIZE  ||
    icon.height               != MIN_ICON_SIZE &&
    icon.height               != MED_ICON_SIZE &&
    icon.height               != MAX_ICON_SIZE  ||
    icon.color_count          != 2 &&
    icon.color_count          != 8 &&
    icon.color_count          != 16
  )
  {
    error_exit("Invalid icon directory entry format");
  }

  // read bitmap info header
  if (fseek(icon_file, icon.image_offset, SEEK_SET) ||
    fread(&bminfo, sizeof(bminfo), 1, icon_file) != 1)
  {
    error_exit("Icon file format error");
  }

  if (verbose)
  {
    printf("bits per pixel     =  %d\n",  bminfo.bits_per_pixel   );
  }

  // check bitmap info header for validity
  if
  (
    bminfo.size                != sizeof(bminfo) ||
    bminfo.width2              != icon.width ||
    bminfo.height2             != 2 * icon.height ||
    bminfo.bits_per_pixel      != 1 &&
    bminfo.bits_per_pixel      != 4 ||
    bminfo.compression         != 0 ||
    bminfo.xpels               != 0 ||
    bminfo.ypels               != 0 ||
    bminfo.colors_used         != 0 ||
    bminfo.important_colors    != 0
  )
  {
    error_exit("Invalid icon header format");
  }

  // read in color table
  if (fread(colors, actual_color_count * sizeof(RGB4), 1, icon_file) != 1)
    error_exit("Icon file does not contain a complete color table");

  // compute length of padded scan line
  scan_line_length =
    ((icon.width * bminfo.bits_per_pixel + 0x1F) & 0xFFE0) / 8;

  // fill in bitmap header
  bitmap_header.type[0] = 'B';
  bitmap_header.type[1] = 'M';
  bitmap_header.file_size = sizeof(bitmap_header) +
    actual_color_count * sizeof(RGB4) + scan_line_length * icon.height;
  bitmap_header.reserved1 = 0;
  bitmap_header.reserved2 = 0;
  bitmap_header.bit_table_offset = sizeof(bitmap_header) +
    actual_color_count * sizeof(RGB4);
  bitmap_header.info_header_size = sizeof(bminfo);
  bitmap_header.width = icon.width;
  bitmap_header.height = icon.height;
  bitmap_header.planes = 1;
  bitmap_header.bits_per_pixel = bminfo.bits_per_pixel;
  bitmap_header.compression = 0;
  bitmap_header.size_of_image = 0;
  bitmap_header.xpels = 0;
  bitmap_header.ypels = 0;
  bitmap_header.colors_used = 0;
  bitmap_header.important_colors = 0;

  // open bitmap file
  bitmap_file = fopen(bitmap_file_specifications,
    append ? "a" : rc_output ? "w" : "wb");
  if (bitmap_file == NULL)
    error_exit("Can't open bitmap file %s", bitmap_file_specifications);

  // add prologue for .RC bitmap
  if (rc_output)
    fprintf(bitmap_file, "\n%s BITMAP\nBEGIN\n", bitmap_name);

  // write bitmap header
  write_bitmap_block(&bitmap_header, sizeof(bitmap_header));

  // write color table
  write_bitmap_block(colors, actual_color_count * sizeof(RGB4));

  // read the XOR colors into the pixel[] array
  {
    unsigned k = 0;
    for (unsigned i = 0; i < icon.height; i++)
    {
      BYTE line[32];
      if (fread(line, scan_line_length, 1, icon_file) != 1)
        error_exit("Icon file format error");
      for (unsigned j = 0; j < icon.width; j++, k++)
      {
        switch (bminfo.bits_per_pixel)
        {
          case 1:
          {
            pixel[k] = colors[line[j>>3] >> 7-(j&7) & 1];
            break;
          }
          case 4:
          {
            pixel[k] = colors[line[j>>1] >> 4*(1-(j&1)) & 0xF];
            break;
          }
        }
      }
    }
  }

  // read the AND bits and process the pixels
  {
    unsigned k = 0;
    for (unsigned i = 0; i < icon.height; i++)
    {
      BYTE line[MAX_ICON_SIZE / 8];
      if (fread(line, icon.width/8, 1, icon_file) != 1)
        error_exit("Icon file format error");
      for (unsigned j = 0; j < icon.width; j++, k++)
      {
        if (line[j>>3] >> 7-(j&7) & 1)
        {
          pixel[k].red   ^= background_color.red  ;
          pixel[k].green ^= background_color.green;
          pixel[k].blue  ^= background_color.blue ;
        }
      }
    }
  }

  // write the new pixels to the bitmap file
  {
    unsigned k = 0;
    for (unsigned i = 0; i < icon.height; i++)
    {
      BYTE line[MAX_ICON_SIZE / 2];
      memset(line, 0, sizeof(line));
      for (unsigned j = 0; j < icon.width; j++, k++)
      {
        unsigned index = 0;
        RGB4 new_pixel = pixel[k];
        unsigned minimum_distance = 3*255 + 1;
        for (unsigned ii = 0; ii < actual_color_count; ii++)
        {
          unsigned distance_of_i;
          int x;
          x = colors[ii].red - new_pixel.red;
          distance_of_i = x < 0 ? -x : x;
          x = colors[ii].green - new_pixel.green;
          distance_of_i += x < 0 ? -x : x;
          x = colors[ii].blue - new_pixel.blue;
          distance_of_i += x < 0 ? -x : x;
          if (distance_of_i < minimum_distance)
          {
            index = ii;
            minimum_distance = distance_of_i;
          }
        }
        switch (bminfo.bits_per_pixel)
        {
          case 1:
          {
            line[j>>3] |= index << 7-(j&7);
            break;
          }
          case 4:
          {
            line[j>>1] |= index << 4*(1-(j&1));
            break;
          }
        }
      }
      write_bitmap_block(line, scan_line_length);
    }
  }

  // For .RC output, flush the buffer and add the required postamble
  if (rc_output)
  {
    flush_bitmap_buffer();
    fprintf(bitmap_file, "END\n");
  }

  fclose(icon_file);
  fclose(bitmap_file);
  return 0;
}

